/*
 *  Filter Object Framework for MICO
 *	Pranav Nabar & Amit Padalkar
 *	8 Jan 2002
 */
#include <string>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cassert>
#include <sys/time.h>
 
#include <CORBA-SMALL.h>
#include <mico/CosNaming.h>
#include "fconfig.h"
#include "lexer.h"

void Config::usage()
{
	cout << "Usage: filterconf [NamingAddress] [Address] [-k key] [-f file] [-n | -nc | -nf] "
		 << "[(-e -d -m) | (-p -u)] [-t] [-a] [-h]\n" << flush;
	cout << "\n\tNamingAddress: Specifies naming service address. \n\t\tThis address must be specified if "
		 <<	"one of the naming options is used.";
	cout << "\n\tAddress: Specifies BOA daemon's address.";
	cout << "\n\t-k key: Use 'key' section in filter configuration file.\n"
		 << "\t\tIf it is not specified, first section in filter configuration file is used.";
	cout << "\n\t-f file: Use 'file' instead of default configuration file 'filter.config'.";
	cout << "\n\t-n | -nc | -nf: '-n' option tells filterconf to use naming service for both client and "
		 << "filter. \n\t\t'-nc' and '-nf' options cause filterconf to use naming service only for client or "
		 << "filter respectively. \n\t\t'NamingAddress' must be specified if these options are used.\n\t\t"
		 << "Moreover BOA daemon's 'Address' might also be requied.";
	cout << "\n\t-e: Use 'Enable' subsection in filter configuration file.";
	cout << "\n\t-d: Use 'Disable' subsection in filter configuration file.";
	cout << "\n\t-m: Use 'Mappings' subsection in filter configuration file.";
	cout << "\n\t-p: Use 'Plug' subsection in filter configuration file.";
	cout << "\n\t-u: Use 'Unplug' subsection in filter configuration file.";
	cout << "\n\t-t: Display time required to send each beta message.";
	cout << "\n\t-a: Use addresses specified in the configuration file.";
	cout << "\n\t-h: Display this message.\n";
	cout << "\nNote:\tIf '-n' option is used along with 'Address', 'Address' will be ignored.\n";
	cout << "\tIf '-a' option is not used, then\n";
	cout << "\t\tIf '-nf' option is used along with '-p' or '-u', 'Address' must be specified.\n";
	cout << "\t\tIf '-nc' option is used, 'Address' must be specified.\n";
	cout << "\t\tIf naming options are not used, 'Address' must be specified.\n";
	cout << "\tIf 'NamingAddress' is specified, it must be first argument followed by others in any order.\n";
	cout << "\tOptions '-e|-d|-m' can not be combined with options '-p|-u'.\n";
}

Config:: Config(int argc, char *argv[]) : file( Lexer::defaultConfig )
{
	isKey = isEnable = isDisable = isMapping = isPlug = isUnplug = isTime = isAddrFromFile = false;
	idStatus = NSNone;

	orb = CORBA::ORB_init( argc, argv, "mico-local-orb" );
	boa = orb->BOA_init (argc, argv, "mico-local-boa");
	
	for( int i = 1; i < argc; ++i )
	{
		string arg( argv[i] );

		if( arg.find(':') != string::npos )
		{
			address = arg;
			continue;
		}
		else if( arg == "-h" )
		{
			usage();
			exit(0);
		}
		else if( arg == "-f" )
		{
			++i;
			if( argv[i][0] == '-' )
			{
				usage();
				exit(1);
			}
			else
				file = argv[i];
		}
		else if( arg == "-k" )
		{
			++i;
			if( argv[i][0] == '-' )
			{
				usage();
				exit(1);
			}
			else
			{
				isKey = true;
				key = argv[i];
			}
		}
		else if( arg[0] == '-' )
		{
			if( arg.length() == 1 )
			{
				usage();
				exit(1);
			}

			for( int j = 1; j < arg.length(); ++j )
			{
				switch( arg[j] )
				{
					case 'e':
						isEnable = true;
						break;
					case 'd':
						isDisable = true;
						break;
					case 'm':
						isMapping = true;
						break;
					case 'p':
						isPlug = true;
						break;
					case 'u':
						isUnplug = true;
						break;
					case 't':
						isTime = true;
						break;
					case 'a':
						isAddrFromFile = true;
						break;
					case 'n':
						if( (arg[j+1] != 'f') && (arg[j+1] != 'c') )
							idStatus = NSAll;
						else
							idStatus = (arg[++j] == 'f') ? NSFilter : NSClient;
						break;
					default:
						usage();
						exit(1);
						break;
				}
			}
		}
		else
		{
			usage();
			exit(1);
		}
	}

	if( (!isAddrFromFile) && ((idStatus == NSNone) || (idStatus == NSClient)) && !address.length() )
	{
		usage();
		exit(1);
	}
	if( (!isAddrFromFile) && (idStatus == NSFilter) && (isPlug || isUnplug) && !address.length() )
	{
		usage();
		exit(1);
	}
	if( (isEnable || isDisable || isMapping) && (isPlug || isUnplug) )
	{
		usage();
		exit(1);
	}
}

void Config::getInfo( vector<string> &values, bool allowNil, bool allowUpDn ) throw(ConfigError)
{
	string val;
	Tokens token;

	try
	{
		while( (token = lexer.getToken(val)) != RBRACE )
		{
			if( token == VALUE )
				values.push_back( val );
			else if( allowNil && (token == NIL) )
				values.push_back( "" );
			else if( allowUpDn && ((token == UP) || (token == DOWN)) )
				values.push_back( ((token == UP) ? Lexer::up : Lexer::down) );
			else
				throw ConfigError();
		}
	}
	catch(FoundEOF)
	{
		throw ConfigError();
	}
}

void Config::getID( string section, vector<string> &id ) throw(ConfigError)
{
	streampos start = lexer.getPosition();
	streampos last = skipSection();
	
	lexer.setPosition( start );

	if( findKey(section) >= last )
		throw ConfigError();

	getInfo( id, true );

	lexer.setPosition( start );
}

string Config::getEnvVal( const string &name )
{
	string val( name ), retVal( name );

	if( name[0] == '$' )
	{
		val.erase(0,1);
		retVal = getenv( val.c_str() );
	}

	return retVal;
}

CORBA::Object_var Config::getObjFromRepo( const vector<string> &id ) throw(ConfigError)
{
	CORBA::Object_var obj;
	int multipleOf = 2;
	string bindAddr;
	int argc = 1;
	char *argv[] = { "filterconf" };

	if( isAddrFromFile )
	{
		++multipleOf;
		orb = CORBA::ORB_init( argc, argv, "mico-local-orb", TRUE );
		boa = orb->BOA_init (argc, argv, "mico-local-boa");
	}

	if( id.size() != multipleOf )
		throw ConfigError();

	if( (!isAddrFromFile) && (address.find(':') == string::npos) )
	{
		cerr << "For using RepoID = " << id[0] << " and Tag = " << id[1] << 
			 ", Address must be specified as first argument.\n" << flush;
		throw ConfigError();
	}

	bindAddr = isAddrFromFile ? getEnvVal(id[2]) : address;
		
	CORBA::ORB::ObjectTag_var tag = CORBA::ORB::string_to_tag ( id[1].c_str() );
	obj = orb->bind ( id[0].c_str(), tag.in(), bindAddr.c_str() );
	assert ( !CORBA::is_nil (obj) );

	return obj;
}

CORBA::Object_var Config::getObjFromNS( const vector<string> &id ) throw(ConfigError)
{
	CORBA::Object_var obj;
	int addElems = isAddrFromFile ? 3 : 1;
	int minSize = isAddrFromFile ? 5 : 3;
	int argc = 5;
	char *argv[] = { "filterconf", "-ORBImplRepoAddr", NULL, "-ORBNamingAddr", NULL };
	char imrAddr[401], namingAddr[401];

	if(  !( (id.size() >= minSize) && !((id.size() - addElems) % 2) )  )
		throw ConfigError();

	istringstream slen( id[0] );
	int len = 0;
	slen >> len;
	if(  !( (len > 0) && (len == ((id.size() - addElems) / 2)) )  )
		throw ConfigError();

	if( isAddrFromFile )
	{
		string imr = getEnvVal( id[id.size()-2] );
		string naming = getEnvVal( id[id.size()-1] );
		strcpy( imrAddr, imr.c_str() );
		strcpy( namingAddr, naming.c_str() );
		argv[2] = imrAddr;
		argv[4] = namingAddr;
		orb = CORBA::ORB_init( argc, argv, "mico-local-orb", TRUE );
		boa = orb->BOA_init (argc, argv, "mico-local-boa");
	}

   	CORBA::Object_var nobj = orb->resolve_initial_references ("NameService");
   	assert (! CORBA::is_nil (nobj));

   	CosNaming::NamingContext_var nc = CosNaming::NamingContext::_narrow (nobj);
   	CosNaming::Name name;
   	name.length( len );
	for( int j = 0; j < len; ++j )
	{
   		name[j].id = CORBA::string_dup ( id[j*2+1].c_str() );
   		name[j].kind = CORBA::string_dup ( id[j*2+2].c_str() );
	}

#ifdef HAVE_EXCEPTIONS
   	try {
       	obj = nc->resolve (name);
   	}
   	catch (CosNaming::NamingContext::NotFound_catch &exc) {
       	cerr << "NotFound exception" << endl;
   	}
   	catch (CosNaming::NamingContext::CannotProceed_catch &exc) {
       	cerr << "CannotProceed exception" << endl;
   	}
   	catch (CosNaming::NamingContext::InvalidName_catch &exc) {
       	cerr << "InvalidName exception" << endl;
   	}
#else
   	obj = nc->resolve (name);
#endif

	assert ( !CORBA::is_nil (obj) );

	return obj;
}

int Config::getRepoId( const vector<string> &ids, int start, vector<string> &id ) throw(ConfigError)
{
	int nextPos, multipleOf = 2;
	
	if( isAddrFromFile )
		++multipleOf;

	if( ids.size() % multipleOf )
		throw ConfigError();

	id.clear();
	for( int i=0; i < multipleOf; ++i )
		id.push_back( ids[start+i] );

#ifdef FCONF_DEBUG
	cout << "Start = " << start << " id = " << id[0] << "#" << id[1];
	if( isAddrFromFile )
		cout << " Addr = " << id[2];
	cout << "	Returning = " << (start+multipleOf) << "\n" << flush;
#endif

	return (start + multipleOf);
}

int Config::getNSId( const vector<string> &ids, int start, vector<string> &id ) throw(ConfigError)
{
	int addElems = isAddrFromFile ? 3 : 1;
	
	id.clear();

	istringstream slen( ids[start] );
	int len = 0;
	slen >> len;
	if(  !( (len > 0) && (ids.size() >= (start+(len*2)+addElems)) )  )
		throw ConfigError();

	for( int i = start; i < (start+(len*2)+addElems); ++i )
		id.push_back( ids[i] );
	
#ifdef FCONF_DEBUG
	cout << "Start = " << start << " id = ";
	for( int x=0; x < id.size(); ++x)
		cout << id[x] << "-|";
	cout << "	Returning = " << (start+(len*2)+addElems) << "\n" << flush;
#endif

	return (start+(len*2)+addElems);
}

void Config::handleClient( const string &section ) throw(ConfigError)
{
	vector<string> filterIds;
	vector<string> clientIds;
	vector<string> filterID;
	vector<string> clientID;
	vector< CORBA::Object_var > clients;
	vector< CORBA::Filter_var > filters;
	CORBA::Object_var obj;
	CORBA::Filter_var fobj;
	int i, j;

	streampos start = lexer.getPosition();
	streampos last = skipSection();
	
	lexer.setPosition( start );
	if( findKey(section) >= last )
		throw ConfigError();

	switch( idStatus )
	{
		case NSNone:
			getID( Lexer::clientRepo, clientIds );
			getID( Lexer::filterRepo, filterIds );

			for( i = 0; i < clientIds.size(); )
			{
				i = getRepoId( clientIds, i , clientID );
				obj = getObjFromRepo( clientID );
				clients.push_back( obj );
			}
			for( i = 0; i < filterIds.size(); )
			{
				i = getRepoId( filterIds, i , filterID );
				obj = getObjFromRepo( filterID );
				fobj = CORBA::Filter::_narrow( obj );
				filters.push_back( fobj );
			}

			break;
		case NSAll:
			getID( Lexer::clientNS, clientIds );
			getID( Lexer::filterNS, filterIds );

			for( i = 0; i < clientIds.size(); )
			{
				i = getNSId( clientIds, i , clientID );
				obj = getObjFromNS( clientID );
				clients.push_back( obj );
			}
			for( i = 0; i < filterIds.size(); )
			{
				i = getNSId( filterIds, i , filterID );
				obj = getObjFromNS( filterID );
				fobj = CORBA::Filter::_narrow( obj );
				filters.push_back( fobj );
			}

			break;
		case NSFilter:
			getID( Lexer::clientRepo, clientIds );
			getID( Lexer::filterNS, filterIds );

			for( i = 0; i < clientIds.size(); )
			{
				i = getRepoId( clientIds, i , clientID );
				obj = getObjFromRepo( clientID );
				clients.push_back( obj );
			}
			for( i = 0; i < filterIds.size(); )
			{
				i = getNSId( filterIds, i , filterID );
				obj = getObjFromNS( filterID );
				fobj = CORBA::Filter::_narrow( obj );
				filters.push_back( fobj );
			}

			break;
		case NSClient:
			getID( Lexer::clientNS, clientIds );
			getID( Lexer::filterRepo, filterIds );

			for( i = 0; i < clientIds.size(); )
			{
				i = getNSId( clientIds, i , clientID );
				obj = getObjFromNS( clientID );
				clients.push_back( obj );
			}
			for( i = 0; i < filterIds.size(); )
			{
				i = getRepoId( filterIds, i , filterID );
				obj = getObjFromRepo( filterID );
				fobj = CORBA::Filter::_narrow( obj );
				filters.push_back( fobj );
			}

			break;
	}

	lexer.setPosition( start );

#ifdef FCONF_DEBUG
	cout << "Client IDs: ";
	for( int x=0; x < clientIds.size(); ++x )
		cout << clientIds[x] << "-|-";
	cout << "\n";
	cout << "Filter IDs: ";
	for( int y=0; y < filterIds.size(); ++y )
		cout << filterIds[y] << "-|-";
	cout << "\n";
#endif

	struct timeval t1,t0;
	long diff;
	int success;
	CORBA::ORB_var orbVar;

	if( section == Lexer::plug )
	{
		for( i=0, j=0; (i < clients.size()) && (j < filters.size()); ++i, ++j )
		{
			success = !gettimeofday(&t0,NULL);
			isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i]->_orbnc();
				orbVar->plug( CORBA::Object_var::duplicate(clients[i]), CORBA::Filter_var::duplicate(filters[j]) );
			}
			else
				orb->plug( CORBA::Object_var::duplicate(clients[i]), CORBA::Filter_var::duplicate(filters[j]) );
			success = !gettimeofday(&t1,NULL);
			isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Plug' = " << diff << "\n" << flush;
			}
		}
		for( ; i < clients.size() ; ++i )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i]->_orbnc();
				orbVar->plug( CORBA::Object_var::duplicate(clients[i]),CORBA::Filter_var::duplicate(filters[j-1]) );
			}
			else
				orb->plug( CORBA::Object_var::duplicate(clients[i]), CORBA::Filter_var::duplicate(filters[j-1]) );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Plug' = " << diff << "\n" << flush;
			}
		}
		for( ; j < filters.size() ; ++j )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i-1]->_orbnc();
				orbVar->plug( CORBA::Object_var::duplicate(clients[i-1]),CORBA::Filter_var::duplicate(filters[j]) );
			}
			else
				orb->plug( CORBA::Object_var::duplicate(clients[i-1]), CORBA::Filter_var::duplicate(filters[j]) );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Plug' = " << diff << "\n" << flush;
			}
		}
	}
	else if( section == Lexer::unplug )
	{
		for( i=0, j=0; (i < clients.size()) && (j < filters.size()); ++i, ++j )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i]->_orbnc();
				orbVar->unplug( CORBA::Object_var::duplicate(clients[i]),CORBA::Filter_var::duplicate(filters[j]) );
			}
			else
				orb->unplug( CORBA::Object_var::duplicate(clients[i]), CORBA::Filter_var::duplicate(filters[j]) );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Unplug' = " << diff << "\n" << flush;
			}
		}
		for( ; i < clients.size() ; ++i )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i]->_orbnc();
				orbVar->unplug(CORBA::Object_var::duplicate(clients[i]),CORBA::Filter_var::duplicate(filters[j-1]));
			}
			else
				orb->unplug( CORBA::Object_var::duplicate(clients[i]), CORBA::Filter_var::duplicate(filters[j-1]) );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Unplug' = " << diff << "\n" << flush;
			}
		}
		for( ; j < filters.size() ; ++j )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			if( isAddrFromFile )
			{
				orbVar = clients[i-1]->_orbnc();
				orbVar->unplug(CORBA::Object_var::duplicate(clients[i-1]),CORBA::Filter_var::duplicate(filters[j]));
			}
			else
				orb->unplug( CORBA::Object_var::duplicate(clients[i-1]), CORBA::Filter_var::duplicate(filters[j]) );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Unplug' = " << diff << "\n" << flush;
			}
		}
	}
}

void Config::handleFilter( const string &section ) throw(ConfigError)
{
	vector<string> filterID;
	vector<string> methods;

	if( (idStatus == NSAll) || (idStatus == NSFilter) )
		getID( Lexer::filterNS, filterID );
	else
		getID( Lexer::filterRepo, filterID );

#ifdef FCONF_DEBUG
	cout << "Filter ID: ";
	for( int y=0; y < filterID.size(); ++y )
		cout << filterID[y] << "-|-";
	cout << "\n";
#endif

	streampos start = lexer.getPosition();
	streampos last = skipSection();
	
	lexer.setPosition( start );
	if( findKey(section) >= last )
		throw ConfigError();

	if( section == Lexer::mappings )
		getInfo( methods, false, true );
	else
		getInfo( methods, false );
	lexer.setPosition( start );

#ifdef FCONF_DEBUG
	cout << "Methods: ";
	for( int x=0; x < methods.size(); ++x )
		cout << methods[x] << "-|-";
	cout << "\n";
#endif

    CORBA::Object_var obj;
	if( (idStatus == NSAll) || (idStatus == NSFilter) )
		obj = getObjFromNS( filterID );
	else
		obj = getObjFromRepo( filterID );

	CORBA::Filter_var filter = CORBA::Filter::_narrow( obj );

	if( section == Lexer::mappings )
	{
		if(  !( (methods.size() >= 3) && !(methods.size() % 3) )  )
			throw ConfigError();
	}
	
	struct timeval t1,t0;
	long diff;
	int success;

	for( int i = 0; i < methods.size(); ++i )
	{
		if( section == Lexer::enable )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			filter->enable( methods[i].c_str() );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Enable' = " << diff << "\n" << flush;
			}
		}
		else if( section == Lexer::disable )
		{
            success = !gettimeofday(&t0,NULL);
            isTime = isTime && success;
			filter->disable( methods[i].c_str() );
            success = !gettimeofday(&t1,NULL);
            isTime = isTime && success;
			if( isTime )
			{
				diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
				cout << "Time(in microsecs) for 'Disable' = " << diff << "\n" << flush;
			}
		}
		else if( section == Lexer::mappings )
		{
			if( methods[i] == Lexer::up )
			{
            	success = !gettimeofday(&t0,NULL);
            	isTime = isTime && success;
				filter->upfilter( methods[i+1].c_str(), methods[i+2].c_str() );
                success = !gettimeofday(&t1,NULL);
                isTime = isTime && success;
				if( isTime )
				{
					diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
					cout << "Time(in microsecs) for 'UpFilter' = " << diff << "\n" << flush;
				}
			}
			else if( methods[i] == Lexer::down )
			{
                success = !gettimeofday(&t0,NULL);
                isTime = isTime && success;
				filter->downfilter( methods[i+1].c_str(), methods[i+2].c_str() );
                success = !gettimeofday(&t1,NULL);
                isTime = isTime && success;
				if( isTime )
				{
					diff = (long)( (t1.tv_sec - t0.tv_sec) * 1E+6 ) + t1.tv_usec - t0.tv_usec;
					cout << "Time(in microsecs) for 'DownFilter' = " << diff << "\n" << flush;
				}
			}
			else
				throw ConfigError();

			i += 2;
		}
	}
}

streampos Config::skipSection() throw(ConfigError)
{
	string val;
	Tokens token;
	int braces = 1;

#ifdef FCONF_DEBUG
	cout << "Trying to skip section. From pos = " << lexer.getPosition() << ".\n" << flush;
	try
	{
		if( lexer.nextToken() == LBRACE )
		{
			cout << "Wrong left brace.\n" << flush;
			throw ConfigError();
		}
	}
	catch( FoundEOF )
	{
		throw ConfigError();
	}
#endif

	do
	{
		try
		{
			token = lexer.getToken(val);

			if( token == LBRACE )
				++braces;
			else if( token == RBRACE )
				--braces;
		}
		catch( FoundEOF )
		{
			throw ConfigError();
		}
	} while( braces > 0 );

#ifdef FCONF_DEBUG
	cout << "Skipped section. Last pos = " << lexer.getPosition() << ".\n" << flush;
#endif

	return lexer.getPosition();
}

bool Config::isSectionKeyword( const string &skey )
{
	return ( 
			(skey == Lexer::plug) || (skey == Lexer::unplug) || (skey == Lexer::enable) || 
			(skey == Lexer::disable) || (skey == Lexer::mappings) || (skey == Lexer::filterRepo) || 
			(skey == Lexer::filterNS) || (skey == Lexer::clientNS) || (skey == Lexer::clientRepo) 
		   );
}

bool Config::isSectionToken( Tokens tok )
{
	return ( 
			(tok == PLUG) || (tok == UNPLUG) || (tok == ENABLE) || 
			(tok == DISABLE) || (tok == MAPPINGS) || (tok == FILTER_REPO) || 
			(tok == FILTER_NS) || (tok == CLIENT_NS) || (tok == CLIENT_REPO) 
		   );
}

streampos Config::findKey(const string &skey) throw(ConfigError)
{
	string val;
	bool keyFound = false;
	bool isKeyword = isSectionKeyword( skey );
	Tokens tok;

#ifdef FCONF_DEBUG
	cout << "Trying to find Key = " << skey << ". From pos = " << lexer.getPosition() << ".\n" << flush;
	cout << "Type of Key = " << (isKeyword ? "SectionKey" : "Value") << ".\n" << flush;
#endif

	do
	{
		try
		{
			if( lexer.getToken(val) != LSQUARE )
				throw ConfigError();

			tok = lexer.getToken(val);
			if( isKeyword )
			{
				if( !isSectionToken(tok) )
					throw ConfigError();
				else
					keyFound = ( tok == lexer.getKeyToken(skey) );
			}
			else
			{
				if( tok != VALUE )
					throw ConfigError();
				else
					keyFound = ( (skey == "") || (val == skey) );
			}

			if( lexer.getToken(val) != RSQUARE )
				throw ConfigError();

			if( lexer.getToken(val) != LBRACE )
				throw ConfigError();

			if( !keyFound )
				skipSection();		
		}
		catch( FoundEOF )
		{
			throw ConfigError();
		}
	} while( !keyFound );

#ifdef FCONF_DEBUG
	cout << "Key = " << skey << ". Found at pos = " << lexer.getPosition() << ".\n" << flush;
#endif

	return lexer.getPosition();
}

void Config:: parse() throw(ConfigError)
{
	try
	{
		lexer.openFile(file);
	}
	catch(CantOpenFile ferr)
	{
		ferr.ShowError();
		throw ConfigError();
	}

	if( isKey )
		findKey( key );
	else
		findKey();

#ifdef FCONF_DEBUG
	cout << "file = " << file << "\n" << flush;
	cout << "addr = " << address << "\n" << flush;
	cout << "isKey = " << isKey << "\n" << flush;
	cout << "Key = " << key << ".\n" << flush;
	cout << "idStatus = " << idStatus << "\n" << flush;
	cout << "isEnable = " << isEnable << "\n" << flush;
	cout << "isDisable = " << isDisable << "\n" << flush;
	cout << "isMapping = " << isMapping << "\n" << flush;
	cout << "isPlug = " << isPlug << "\n" << flush;
	cout << "isUnplug = " << isUnplug << "\n" << flush;
	cout << "isTime = " << isTime << "\n" << flush;
	cout << "isAddrFromFile = " << isAddrFromFile << "\n" << flush;
#endif

	if( isMapping )
		handleFilter( Lexer::mappings );

	if( isEnable )
		handleFilter( Lexer::enable );

	if( isDisable )
		handleFilter( Lexer::disable );

	if( isPlug )
		handleClient( Lexer::plug );

	if( isUnplug )
		handleClient( Lexer::unplug );
}

