/*
 *  Filter Object Framework for MICO
 *	Pranav Nabar & Amit Padalkar
 *	8 Jan 2002
 */
 
#include <iostream>
#include <fstream>
#include <string>
#include <map>

#include "lexer.h"

map< string, Tokens > Lexer:: keywords;
map< string, Tokens > Lexer:: punct;
const string Lexer:: defaultConfig( "filter.config" );
const string Lexer:: plug( "Plug" );
const string Lexer:: unplug( "Unplug" );
const string Lexer:: enable( "Enable" );
const string Lexer:: disable( "Disable" );
const string Lexer:: mappings( "Mappings" );
const string Lexer:: filterRepo( "FilterRepo" );
const string Lexer:: filterNS( "FilterNS" );
const string Lexer:: clientRepo( "ClientRepo" );
const string Lexer:: clientNS( "ClientNS" );
const string Lexer:: up( "Up" );
const string Lexer:: down( "Down" );
const string Lexer:: nil( "Nil" );
const string Lexer:: lbrace( "{" );
const string Lexer:: rbrace( "}" );
const string Lexer:: lsquare( "[" );
const string Lexer:: rsquare( "]" );

void Lexer :: initTokens(void)
{
	if( punct.empty() )
	{
 		keywords[plug] = PLUG;
		keywords[unplug] = UNPLUG;
   		keywords[enable] = ENABLE;
   		keywords[disable] = DISABLE;
   		keywords[mappings] = MAPPINGS;
   		keywords[filterRepo] = FILTER_REPO;
   		keywords[filterNS] = FILTER_NS;
   		keywords[clientRepo] = CLIENT_REPO;
   		keywords[clientNS] = CLIENT_NS;
   		keywords[up] = UP;
   		keywords[down] = DOWN;
   		keywords[nil] = NIL;
		
   		punct[lbrace] = LBRACE;
   		punct[rbrace] = RBRACE;
   		punct[lsquare] = LSQUARE;
   		punct[rsquare] = RSQUARE;
	}
}

void Lexer :: skipSpace(void) const throw(FoundEOF)
{
	char ch;
	
	currFile.seekg(start_pos);
	while( currFile.get(ch) )
	{
		if( (ch != ' ') && (ch != '\t') && (ch != '\n') )
			break;
		else
			start_pos = currFile.tellg();
	}

	if( currFile.eof() )
	{
		throw FoundEOF(filename);
	}

}

bool Lexer :: comment(void) const throw(FoundEOF,UnterminatedComment)
{
	char prev, curr = '/';
	bool ret_val;
	
	currFile.seekg(start_pos);
	if( !currFile.get(prev) )
	{
		throw FoundEOF(filename);
	}
	if( !currFile.get(curr) )
	{
		ret_val = false;
		ClearFState();
	}

	if( (prev == '/') && (curr == '*') )
	{
		curr = prev;
		do
		{
			prev = curr;
			if( !currFile.get(curr) )
			{
				throw UnterminatedComment(start_pos);
			}
		}
		while( !( (prev == '*') && (curr == '/') ) );
		start_pos = currFile.tellg();
		ret_val = true;
	}
	else
		ret_val = false;

	skipSpace();
	return ret_val;
}

Tokens Lexer :: punctuation(void) const
{
	char ch;
	Tokens ret_val;
	string val;
	
	currFile.seekg(start_pos);
	
	currFile.get(ch);
	val += ch;
	if( punct.find(val) != punct.end() )
	{
		ret_val = punct[val];
		start_pos = currFile.tellg();
	}
	else
	{
		ret_val = INVALID;
	}

	return ret_val;
}

string Lexer :: getWord(void) const
{
	char ch;
	string ret_val, val;
	
	currFile.seekg(start_pos);
	
	currFile.get(ch);
	val = ch;
	if( (punct.find(val) == punct.end())  && (ch != ' ' )
	    && (ch != '\t' ) && (ch != '\n' ) )
	{
		ret_val += ch;
		start_pos = currFile.tellg();
		while( currFile.get(ch) )
		{
			val = ch;
			if( (punct.find(val) != punct.end())  || (ch == ' ' )
	    		    || (ch == '\t' ) || (ch == '\n' )  )
				break;
			else
			{
				ret_val += ch;
				start_pos = currFile.tellg();
			}	
		}
	}
	
	return ret_val;
}

Tokens Lexer :: getKeyToken(const string &word) const
{
	map<string , Tokens>::iterator i = keywords.find(word);

	if( i == keywords.end() )
		return INVALID;
	else
		return (*i).second;
}

Tokens Lexer :: keyword(void) const
{
	string word;
	Tokens curr_token;
	
	streampos curr_pos = getPosition();
	if( (word = getWord()).length() )
	{
		if( keywords.find(word) != keywords.end() )
		{
			curr_token = keywords[word];
		}
		else
		{
			curr_token = INVALID;
			start_pos = curr_pos;	
		}	
	}
	else
		curr_token = INVALID;

	return curr_token;	
}

Tokens Lexer :: info(string& value)  const
{
	string word;
	Tokens curr_token;
	
	if( (word = getWord()).length() )
	{
		curr_token = VALUE;
	}
	else
	{
		curr_token = INVALID;
	}

	value = word;
	return curr_token;	
}

Lexer :: Lexer (streampos start_pos) : start_pos(start_pos) 
{
	initTokens() ;
}

Lexer :: Lexer (string filename , streampos start_pos) 
throw (CantOpenFile) : filename(filename) , start_pos(start_pos) 
{ 
	currFile.open(filename.c_str()) ;
	
	if( currFile.fail() )
	{
		throw CantOpenFile(filename) ;
	}
	
	initTokens() ;
}

Lexer :: ~Lexer()
{
	currFile.close() ;
}

void Lexer:: openFile(string filename) throw(CantOpenFile)
{
	//currFile.close();
	currFile.open(filename.c_str()) ;
	
	if( currFile.fail() )
	{
		throw CantOpenFile(filename) ;
	}

	this->filename = filename;
	setPosition(0);
}

Tokens Lexer :: getToken(string &value)  const
	throw(FoundEOF, ConfigError)
{
	Tokens curr_token;

	if( currFile.eof() )
	{
		throw FoundEOF(filename);
	}

	try
	{
		skipSpace();
		while( comment() ) ;

		if( (curr_token = punctuation()) != INVALID )
		{
			value.erase();
		}
		else if( (curr_token = keyword()) != INVALID )
		{
			value.erase();
		}
		else if( (curr_token = info(value)) != INVALID )
		{
		}
		else
		{
			value.erase();
			curr_token = INVALID;
		}
	}
	catch( const LexerError& fail )
	{
		fail.ShowError();
		throw ConfigError();
	}

	return curr_token;
}

Tokens Lexer :: nextToken(string& value)  const
	throw(FoundEOF, ConfigError)
{
	streampos current;
	Tokens currToken;
	
	current = getPosition();
	currToken = getToken(value);
	setPosition(current);

	return currToken;
}

Tokens Lexer :: nextToken(void)  const
	throw(FoundEOF, ConfigError)
{
	string value;
	
	return nextToken(value);
}

