/*
 *  Filter Object Framework for MICO
 *	Pranav Nabar & Amit Padalkar
 *	8 Jan 2002
 */

#ifndef LEXER_H

#define LEXER_H

#include <iostream>
#include <fstream>
#include <string>
#include <map>

/*******************************************************************
	Exception hierarchy : 
	--------------------------

	Error
		ConfigError
		FileError
			CantOpenFile
			FoundEOF
		LexerError
			UnterminatedComment

*******************************************************************/

class Error
{
public :
	virtual void ShowError() const = 0;
};

class	ConfigError : public Error
{
public :
	void ShowError()  const	{}
};

class	FileError : public Error
{
protected :
	string filename;

public :
	FileError(string filename) : filename(filename)	{}
	virtual void ShowError() const = 0;
};

class	CantOpenFile : public FileError
{
public :
	CantOpenFile(string filename) : FileError(filename)	{}
	void ShowError() const
	{
		cerr << endl << "Unable to open file : " << filename << endl;
	}
};
	
class	FoundEOF : public FileError
{
public :
	FoundEOF(string filename) : FileError(filename)	{}
	void ShowError() const
	{
		cerr << endl << "Encountered end of file : " << filename <<endl;
	}
};
	
class	LexerError : public Error
{
protected :
	streampos position;
	
public :
	LexerError(streampos position) : position(position)	{}
	virtual void ShowError() const = 0;
};

class	UnterminatedComment : public LexerError
{
public :
	UnterminatedComment(streampos position) 
		: LexerError(position)	{}
	void ShowError() const
	{
		cerr << endl << "Unterminated comment starting at position : " << position << endl;
	}
};
	
enum Tokens
{
	INVALID = -1	,
	LBRACE = 1		,	// Value = 1
	RBRACE			,	// Value = 2
	LSQUARE			,	// Value = 3
	RSQUARE			,	// Value = 4
	COMMENT			,	// Value = 5
	PLUG			,	// Value = 6
	UNPLUG			,	// Value = 7
	ENABLE			,	// Value = 8
	DISABLE			,	// Value = 9
	MAPPINGS		,	// Value = 10
	FILTER_REPO		,	// Value = 11
	FILTER_NS		,	// Value = 12
	CLIENT_REPO		,	// Value = 13
	CLIENT_NS		,	// Value = 14
	UP				,	// Value = 15
	DOWN			,	// Value = 16
	VALUE			,	// Value = 17
	NIL					// Value = 18
};

class Lexer 
{
private :
	string filename ;
	mutable ifstream currFile ;
	mutable streampos start_pos ;

	static map< string , Tokens > keywords ;
	static map< string , Tokens > punct ;
	
	static void initTokens(void) ;

	void skipSpace(void) const throw(FoundEOF) ;
	bool comment(void) const throw(FoundEOF,UnterminatedComment) ;
	Tokens punctuation(void) const;
	string getWord(void) const ;
	Tokens keyword(void)  const;
	Tokens info(string& value) const ;

protected :

public :
	static const string defaultConfig ;
	static const string plug;
	static const string unplug;
	static const string enable;
	static const string disable;
	static const string mappings;
	static const string filterRepo;
	static const string filterNS;
	static const string clientRepo;
	static const string clientNS;
	static const string up;
	static const string down;
	static const string nil;
	static const string lbrace;
	static const string rbrace;
	static const string lsquare;
	static const string rsquare;


	Lexer(streampos start_pos = 0);
	Lexer(string filename, streampos start_pos = 0) throw(CantOpenFile) ;
	~Lexer() ;

	void ClearFState(void) const
	{
		currFile.clear();
	}

	streampos getPosition(void)  const
	{
		return start_pos ;
	}
	
	void setPosition(streampos start_pos)  const
	{
		this->start_pos = start_pos ;
	}

	void openFile(string filename) throw(CantOpenFile);
	Tokens getToken(string& value)  const throw(FoundEOF, ConfigError);
	Tokens nextToken(string& value)  const throw(FoundEOF, ConfigError);
	Tokens nextToken(void)  const throw(FoundEOF, ConfigError);
	Tokens getKeyToken(const string &word) const;
};

#endif

