/*
 *  Filter Object Framework for MICO
 *	Pranav Nabar & Amit Padalkar
 *	8 Jan 2002
 */

#ifndef _FILTER_H
#define _FILTER_H


#ifndef FILTER_DEBUG
//	#define FILTER_DEBUG    // Uncomment this line to get Filter Framework debug messages
#endif


class Filter;
typedef Filter *Filter_ptr;
typedef Filter_ptr FilterRef;
typedef ObjVar<Filter> Filter_var;
typedef ObjOut<Filter> Filter_out;
typedef vector< Filter_var > FilterVec;

/*
	UpDownMethods structure is used to store names of 'upfilter' and 'downfilter' methods.
*/
struct UpDownMethods
{
	public:
		vector<string> upFilter;
		vector<string> downFilter;
};

/*
	MethodMappings stores mappings between filter client methods & corresponding 'upfilter' & 'downfilter' methods.
*/
typedef map< string, UpDownMethods > MethodMappings;	
typedef map< string, UpDownMethods >::value_type MMValType;	

/*
	FilterMethodStatusMap stores enable/disable status of all filter methods.
*/
typedef map< string, bool > FilterMethodStatusMap;	
typedef map< string, bool >::value_type FMSValType;	

class FilterServerRequest;

/* Base class of all filter objects */
class Filter : virtual public CORBA::Object
{
	private:
		FilterMethodStatusMap filterMethods;	// Status of every filter method.
		MethodMappings mappings;	// Mappings of filter client methods to corresponding UpDown filter methods.
		CORBA::Boolean passFilterClientMethod;	// Pass/Bounce status of filter client method.

		void restoreFilter();

		const FilterMethodStatusMap& statusInfo()
		{
			return filterMethods;
		}

		const MethodMappings& mappingInfo()
		{
			return mappings;
		}

		CORBA::Boolean getPassStatus()
		{
			return passFilterClientMethod;
		}

		CORBA::Boolean& getClientPassRef()
		{
			return passFilterClientMethod;
		}

		friend class FilterServerRequest;

	public:
		static Filter_ptr _narrow( CORBA::Object_ptr obj );

    	static Filter_ptr _duplicate( Filter_ptr _obj )
    	{
      		CORBA::Object::_duplicate (_obj);
     	 	return _obj;
    	}
 
    	static Filter_ptr _nil()
    	{
      		return 0;
    	}

		void saveFilter();

		/* Interface for enabling and disabling appropriate filter methods. */
		void enable( string method, CORBA::Boolean isFilterSkel = FALSE );
		void disable( string method, CORBA::Boolean isFilterSkel = FALSE );

		/* Interface for adding a mapping specifying filter client method and appropriate Up/Down filter method. */
		void upfilter( string clientMethod, string filterMethod, CORBA::Boolean isFilterSkel = FALSE );
		void downfilter( string clientMethod, string filterMethod, CORBA::Boolean isFilterSkel = FALSE );

		void setPass()
		{
			passFilterClientMethod = TRUE;
		}

		void setBounce()
		{
			passFilterClientMethod = FALSE;
		}
};


/*
	MethodStatus indicates state of arrived method call.
*/
enum MethodStatus
{
	ServerNormal,	// Method call at server object that is not a filter.
	FilterNormal,	// Direct method call at filter object.
	FilterClientNormal,		// Intercepted method call at filter object.
	FilterClientBad		// Intercepted method call at filter but without proper mapping or disabled filter method.
};

class FilterRequest;

/* Server side filter request. */
class FilterServerRequest : public CORBA::StaticServerRequest
{
	private:
		CORBA::Boolean isFilter;	// Indicates whether object, to which this request is targetted, is filter. 
		CORBA::Boolean passMethod;	// Decides whether requested method is called.
		CORBA::Object_ptr skelObj;
		Filter_var filterObj;
		MethodStatus status;

		CORBA::Boolean doUpFiltering();		// Handles upward filtering.
		void doDownFiltering();		// Handles downward filtering.

		friend class FilterRequest;		// Allow client side filter request access to server side request's members.

	public:
	    FilterServerRequest ( ORBRequest *req, Object_ptr obj, MsgId msgid,
             				  CORBA::ObjectAdapter *oa, CORBA::Principal_ptr pr, CORBA::ImplementationBase *skel );
		~FilterServerRequest()	{}

		const char *op_name();
		CORBA::Boolean read_args();
		void write_results();

		CORBA::Boolean filterStatus()
		{
			return isFilter;
		}
		
		MethodStatus methodStatus()
		{
			return status;
		}
};


/* Client side filter request. */
class FilterRequest : public CORBA::StaticRequest
{
	private:
		CORBA::Boolean isUpFilter;	// Indicates whether up/down filtering is desired for requested method.
		CORBA::String_var oldOpName;	// Method name before up/down transformation.

	public:
		FilterRequest( Filter_ptr, const char *opn, FilterServerRequest *fsreq, 
					   CORBA::Boolean isUp = TRUE, CORBA::Boolean createSRIcep = FALSE );
		~FilterRequest()	{}
};


/*
	BetaMessage class represents a message intended for filter framework components rather than to any
	specific object. Its interface has single 'handle' method intended to handle the corresponding beta message.
*/
class BetaMessage
{
protected:
	const char *opName;
	CORBA::ImplementationBase *skel;
	CORBA::ORBRequest *req;
	CORBA::Object_ptr obj;
	CORBA::ULong msgid;
	CORBA::ObjectAdapter *oa;
	CORBA::Principal_ptr pr;

public:
	BetaMessage( CORBA::ImplementationBase *skel, CORBA::ORBRequest *req, CORBA::Object_ptr obj, 
				 CORBA::ULong msgid, CORBA::ObjectAdapter *oa, CORBA::Principal_ptr pr );
	virtual ~BetaMessage()	{ }

	 virtual CORBA::Boolean handle( ) = 0;
};

/*
	PlugUnplugMessage is a beta message. Its handling involves plugging or unplugging appropriate filter object
	from proper filter client using FilterMapper.
*/
class PlugUnplugMessage : public BetaMessage
{
public:
	PlugUnplugMessage( CORBA::ImplementationBase *skel, CORBA::ORBRequest *req, CORBA::Object_ptr obj, 
				 	   CORBA::ULong msgid, CORBA::ObjectAdapter *oa, CORBA::Principal_ptr pr );
	~PlugUnplugMessage()	{ }

	 CORBA::Boolean handle( );
};

/*
	EnableDisableMessage is a beta message. Its handling involves enabling or disabling filter method
	in appropriate filter object.
*/
class EnableDisableMessage : public BetaMessage
{
public:
	EnableDisableMessage( CORBA::ImplementationBase *skel, CORBA::ORBRequest *req, CORBA::Object_ptr obj, 
				 		  CORBA::ULong msgid, CORBA::ObjectAdapter *oa, CORBA::Principal_ptr pr );
	~EnableDisableMessage()	{ }

	 CORBA::Boolean handle( );
};

/*
	UpDownFilterMessage is a beta message. Its handling involves adding mapping between filter client method 
	and up/down filter method in appropriate filter object.
*/
class UpDownFilterMessage : public BetaMessage
{
public:
	UpDownFilterMessage( CORBA::ImplementationBase *skel, CORBA::ORBRequest *req, CORBA::Object_ptr obj, 
				 		  CORBA::ULong msgid, CORBA::ObjectAdapter *oa, CORBA::Principal_ptr pr );
	~UpDownFilterMessage()	{ }

	 CORBA::Boolean handle( );
};


#endif	// _FILTER_H

